@@ -11,6 +11,8 @@ use camino::{Utf8Path, Utf8PathBuf};
 use clap::{ArgEnum, Args, Parser, Subcommand};
 use color_eyre::eyre::{Report, Result, WrapErr};
 use guppy::graph::PackageGraph;
+use itertools::Itertools;
+use nextest_filtering::FilteringExpr;
 use nextest_metadata::{BinaryListSummary, BuildPlatform};
 use nextest_runner::{
     config::{NextestConfig, NextestProfile},
@@ -285,6 +287,15 @@ struct TestBuildFilter {
     #[clap(long, arg_enum, value_name = "PLATFORM", default_value_t)]
     pub(crate) platform_filter: PlatformFilterOpts,
 
+    /// Test filter expression (see {n}<https://nexte.st/book/filter-expressions>)
+    #[clap(
+        long,
+        short = 'E',
+        value_name = "EXPRESSION",
+        multiple_occurrences(true)
+    )]
+    filter_expr: Vec<String>,
+
     // TODO: add regex-based filtering in the future?
     /// Test name filter
     #[clap(name = "FILTERS", help_heading = None)]
@@ -298,6 +309,7 @@ impl TestBuildFilter {
         binary_list: BinaryList,
         runner: &TargetRunner,
         reuse_build: &ReuseBuildOpts,
+        filter_exprs: Vec<FilteringExpr>,
     ) -> Result<TestList<'g>> {
         let path_mapper =
             reuse_build.make_path_mapper(graph, &binary_list.rust_build_meta.target_directory);
@@ -308,8 +320,12 @@ impl TestBuildFilter {
             &path_mapper,
             self.platform_filter.into(),
         )?;
-        let test_filter =
-            TestFilterBuilder::new(self.run_ignored, self.partition.clone(), &self.filter);
+        let test_filter = TestFilterBuilder::new(
+            self.run_ignored,
+            self.partition.clone(),
+            &self.filter,
+            filter_exprs,
+        );
         TestList::new(
             test_artifacts,
             &rust_build_meta,
@@ -461,6 +477,19 @@ struct App {
     config_opts: ConfigOpts,
 }
 
+fn check_experimental_filtering(build_filter: &TestBuildFilter) -> Result<()> {
+    const EXPERIMENTAL_ENV: &str = "NEXTEST_EXPERIMENTAL_FILTER_EXPR";
+    let enabled = std::env::var(EXPERIMENTAL_ENV).is_ok();
+    if !build_filter.filter_expr.is_empty() && !enabled {
+        Err(Report::new(ExpectedError::experimental_feature_error(
+            "expression filtering",
+            EXPERIMENTAL_ENV,
+        )))
+    } else {
+        Ok(())
+    }
+}
+
 impl App {
     fn new(
         output: OutputOpts,
@@ -470,17 +499,29 @@ impl App {
         manifest_path: Option<Utf8PathBuf>,
     ) -> Result<Self> {
         let output = output.init();
-        reuse_build.check_experimental(output)?;
+        reuse_build.check_experimental(output);
+        check_experimental_filtering(&build_filter)?;
 
         let graph_data = match reuse_build.cargo_metadata.as_deref() {
-            Some(path) => std::fs::read_to_string(path)?,
-            None => acquire_graph_data(
-                manifest_path.as_deref(),
-                build_filter.cargo_options.target_dir.as_deref(),
-                output,
-            )?,
+            Some(path) => std::fs::read_to_string(path).map_err(|err| {
+                ExpectedError::argument_file_read_error("cargo-metadata", path, err)
+            })?,
+            None => {
+                let with_deps = build_filter
+                    .filter_expr
+                    .iter()
+                    .any(|expr| FilteringExpr::needs_deps(expr));
+                acquire_graph_data(
+                    manifest_path.as_deref(),
+                    build_filter.cargo_options.target_dir.as_deref(),
+                    output,
+                    with_deps,
+                )?
+            }
         };
-        let graph = guppy::CargoMetadata::parse_json(&graph_data)?.build_graph()?;
+        let graph = PackageGraph::from_json(&graph_data).map_err(|err| {
+            ExpectedError::cargo_metadata_parse_error(reuse_build.cargo_metadata.clone(), err)
+        })?;
 
         let manifest_path = if reuse_build.cargo_metadata.is_some() {
             Some(graph.workspace().root().join("Cargo.toml"))
@@ -507,8 +548,13 @@ impl App {
     fn build_binary_list(&self) -> Result<BinaryList> {
         let binary_list = match self.reuse_build.binaries_metadata.as_deref() {
             Some(path) => {
-                let raw_binary_list = std::fs::read_to_string(path)?;
-                let binary_list: BinaryListSummary = serde_json::from_str(&raw_binary_list)?;
+                let raw_binary_list = std::fs::read_to_string(path).map_err(|err| {
+                    ExpectedError::argument_file_read_error("binaries-metadata", path, err)
+                })?;
+                let binary_list: BinaryListSummary = serde_json::from_str(&raw_binary_list)
+                    .map_err(|err| {
+                        ExpectedError::argument_json_parse_error("binaries-metadata", path, err)
+                    })?;
                 BinaryList::from_summary(binary_list)
             }
             None => self.build_filter.compute_binary_list(
@@ -520,16 +566,33 @@ impl App {
         Ok(binary_list)
     }
 
+    fn build_filtering_expressions(&self) -> Result<Vec<FilteringExpr>> {
+        let (exprs, all_errors): (Vec<_>, Vec<_>) = self
+            .build_filter
+            .filter_expr
+            .iter()
+            .map(|input| FilteringExpr::parse(input, &self.graph))
+            .partition_result();
+
+        if !all_errors.is_empty() {
+            Err(ExpectedError::filter_expression_parse_error(all_errors).into())
+        } else {
+            Ok(exprs)
+        }
+    }
+
     fn build_test_list(
         &self,
         binary_list: BinaryList,
         target_runner: &TargetRunner,
+        filter_exprs: Vec<FilteringExpr>,
     ) -> Result<TestList> {
         self.build_filter.compute_test_list(
             &self.graph,
             binary_list,
             target_runner,
             &self.reuse_build,
+            filter_exprs,
         )
     }
 
@@ -559,6 +622,7 @@ impl App {
         list_type: ListType,
         output_writer: &mut OutputWriter,
     ) -> Result<()> {
+        let filter_exprs = self.build_filtering_expressions()?;
         let binary_list = self.build_binary_list()?;
 
         match list_type {
@@ -573,7 +637,7 @@ impl App {
             }
             ListType::Full => {
                 let target_runner = self.load_runner();
-                let test_list = self.build_test_list(binary_list, &target_runner)?;
+                let test_list = self.build_test_list(binary_list, &target_runner, filter_exprs)?;
 
                 let mut writer = output_writer.stdout_writer();
                 test_list.write(
@@ -602,8 +666,9 @@ impl App {
 
         let target_runner = self.load_runner();
 
+        let filter_exprs = self.build_filtering_expressions()?;
         let binary_list = self.build_binary_list()?;
-        let test_list = self.build_test_list(binary_list, &target_runner)?;
+        let test_list = self.build_test_list(binary_list, &target_runner, filter_exprs)?;
 
         let mut reporter = reporter_opts
             .to_builder(no_capture)
@@ -634,11 +699,14 @@ fn acquire_graph_data(
     manifest_path: Option<&Utf8Path>,
     target_dir: Option<&Utf8Path>,
     output: OutputContext,
+    with_deps: bool,
 ) -> Result<String> {
     let mut cargo_cli = CargoCli::new("metadata", manifest_path, output);
-    // Construct a package graph with --no-deps since we don't need full dependency
-    // information.
-    cargo_cli.add_args(["--format-version=1", "--all-features", "--no-deps"]);
+    cargo_cli.add_args(["--format-version=1", "--all-features"]);
+
+    if !with_deps {
+        cargo_cli.add_arg("--no-deps");
+    }
 
     let mut expression = cargo_cli.to_expression().stdout_capture().unchecked();
     // cargo metadata doesn't support "--target-dir" but setting the environment
