@@ -17,8 +17,14 @@ mod cargo_cli;
 mod dispatch;
 mod errors;
 mod output;
+mod reuse_build;
+
+#[cfg(test)]
+mod tests_integration;
 
 #[doc(hidden)]
 pub use dispatch::*;
 #[doc(hidden)]
 pub use errors::*;
+#[doc(hidden)]
+pub use output::OutputWriter;
