@@ -1,15 +1,14 @@
 // Copyright (c) The nextest Contributors
 // SPDX-License-Identifier: MIT OR Apache-2.0
 
-use crate::{output::OutputContext, ExpectedError};
+use crate::output::OutputContext;
 use camino::{Utf8Path, Utf8PathBuf};
 use clap::Args;
-use color_eyre::eyre::{Report, Result};
 use guppy::graph::PackageGraph;
 use nextest_runner::list::PathMapper;
 
 #[derive(Debug, Args)]
-#[clap(next_help_heading = "REUSE BUILD OPTIONS (EXPERIMENTAL)")]
+#[clap(next_help_heading = "REUSE BUILD OPTIONS")]
 pub(crate) struct ReuseBuildOpts {
     /// Path to binaries-metadata JSON
     #[clap(long, value_name = "PATH")]
@@ -33,21 +32,9 @@ impl ReuseBuildOpts {
 
     // (_output is not used, but must be passed in to ensure that the output is properly initialized
     // before calling this method)
-    pub(crate) fn check_experimental(&self, _output: OutputContext) -> Result<()> {
-        let used = self.binaries_metadata.is_some()
-            || self.target_dir_remap.is_some()
-            || self.cargo_metadata.is_some()
-            || self.workspace_remap.is_some();
-
-        let enabled = std::env::var(Self::EXPERIMENTAL_ENV).is_ok();
-
-        if used && !enabled {
-            Err(Report::new(ExpectedError::experimental_feature_error(
-                "build reuse",
-                Self::EXPERIMENTAL_ENV,
-            )))
-        } else {
-            Ok(())
+    pub(crate) fn check_experimental(&self, _output: OutputContext) {
+        if std::env::var(Self::EXPERIMENTAL_ENV).is_ok() {
+            log::warn!("build reuse is no longer experimental: NEXTEST_EXPERIMENTAL_REUSE_BUILD does not need to be set");
         }
     }
 
